@@ -2,65 +2,51 @@ package doublestar
 
 import (
 	"fmt"
+	"io"
 	"os"
 	"path"
 	"path/filepath"
+	"sort"
 	"strings"
 	"unicode/utf8"
 )
 
+// File defines a subset of file operations
+type File interface {
+	io.Closer
+	Readdir(count int) ([]os.FileInfo, error)
+}
+
+// An OS abstracts functions in the standard library's os package.
+type OS interface {
+	Lstat(name string) (os.FileInfo, error)
+	Open(name string) (File, error)
+	PathSeparator() rune
+	Stat(name string) (os.FileInfo, error)
+}
+
+// A standardOS implements OS by calling functions in the standard library's os
+// package.
+type standardOS struct{}
+
+func (standardOS) Lstat(name string) (os.FileInfo, error) { return os.Lstat(name) }
+func (standardOS) Open(name string) (File, error)         { return os.Open(name) }
+func (standardOS) PathSeparator() rune                    { return os.PathSeparator }
+func (standardOS) Stat(name string) (os.FileInfo, error)  { return os.Stat(name) }
+
+// StandardOS is a value that implements the OS interface by calling functions
+// in the standard libray's os package.
+var StandardOS OS = standardOS{}
+
 // ErrBadPattern indicates a pattern was malformed.
 var ErrBadPattern = path.ErrBadPattern
 
-// Split a path on the given separator, respecting escaping.
-func splitPathOnSeparator(path string, separator rune) (ret []string) {
-	idx := 0
-	if separator == '\\' {
-		// if the separator is '\\', then we can just split...
-		ret = strings.Split(path, string(separator))
-		idx = len(ret)
-	} else {
-		// otherwise, we need to be careful of situations where the separator was escaped
-		cnt := strings.Count(path, string(separator))
-		if cnt == 0 {
-			return []string{path}
-		}
-
-		ret = make([]string, cnt+1)
-		pathlen := len(path)
-		separatorLen := utf8.RuneLen(separator)
-		emptyEnd := false
-		for start := 0; start < pathlen; {
-			end := indexRuneWithEscaping(path[start:], separator)
-			if end == -1 {
-				emptyEnd = false
-				end = pathlen
-			} else {
-				emptyEnd = true
-				end += start
-			}
-			ret[idx] = path[start:end]
-			start = end + separatorLen
-			idx++
-		}
-
-		// If the last rune is a path separator, we need to append an empty string to
-		// represent the last, empty path component. By default, the strings from
-		// make([]string, ...) will be empty, so we just need to icrement the count
-		if emptyEnd {
-			idx++
-		}
-	}
-
-	return ret[:idx]
-}
-
 // Find the first index of a rune in a string,
 // ignoring any times the rune is escaped using "\".
 func indexRuneWithEscaping(s string, r rune) int {
 	end := strings.IndexRune(s, r)
-	if end == -1 {
-		return -1
+	if end == -1 || r == '\\' {
+		return end
 	}
 	if end > 0 && s[end-1] == '\\' {
 		start := end + utf8.RuneLen(r)
@@ -198,7 +184,7 @@ func isZeroLengthPattern(pattern string) (ret bool, err error) {
 // is the PathMatch() function below.
 //
 func Match(pattern, name string) (bool, error) {
-	return matchWithSeparator(pattern, name, '/')
+	return doMatching(pattern, name, '/')
 }
 
 // PathMatch is like Match except that it uses your system's path separator.
@@ -209,83 +195,74 @@ func Match(pattern, name string) (bool, error) {
 // Note: this is meant as a drop-in replacement for filepath.Match().
 //
 func PathMatch(pattern, name string) (bool, error) {
+	return PathMatchOS(StandardOS, pattern, name)
+}
+
+// PathMatchOS is like PathMatch except that it uses vos's path separator.
+func PathMatchOS(vos OS, pattern, name string) (bool, error) {
 	pattern = filepath.ToSlash(pattern)
-	return matchWithSeparator(pattern, name, os.PathSeparator)
+	return doMatching(pattern, name, vos.PathSeparator())
 }
 
-// Match returns true if name matches the shell file name pattern.
-// The pattern syntax is:
-//
-//  pattern:
-//    { term }
-//  term:
-//    '*'         matches any sequence of non-path-separators
-//              '**'        matches any sequence of characters, including
-//                          path separators.
-//    '?'         matches any single non-path-separator character
-//    '[' [ '^' ] { character-range } ']'
-//          character class (must be non-empty)
-//    '{' { term } [ ',' { term } ... ] '}'
-//    c           matches character c (c != '*', '?', '\\', '[')
-//    '\\' c      matches character c
-//
-//  character-range:
-//    c           matches character c (c != '\\', '-', ']')
-//    '\\' c      matches character c, unless separator is '\\'
-//    lo '-' hi   matches character c for lo <= c <= hi
-//
-// Match requires pattern to match all of name, not just a substring.
-// The only possible returned error is ErrBadPattern, when pattern
-// is malformed.
-//
-func matchWithSeparator(pattern, name string, separator rune) (bool, error) {
-	nameComponents := splitPathOnSeparator(name, separator)
-	return doMatching(pattern, nameComponents)
-}
-
-func doMatching(pattern string, nameComponents []string) (matched bool, err error) {
+func doMatching(pattern, name string, separator rune) (matched bool, err error) {
 	// check for some base-cases
-	patternLen, nameLen := len(pattern), len(nameComponents)
-	if patternLen == 0 && nameLen == 0 {
-		return true, nil
-	}
-	if patternLen == 0 || nameLen == 0 {
-		return false, nil
+	patternLen, nameLen := len(pattern), len(name)
+	if patternLen == 0 {
+		return nameLen == 0, nil
+	} else if nameLen == 0 {
+		return isZeroLengthPattern(pattern)
 	}
 
-	slashIdx := indexRuneWithEscaping(pattern, '/')
-	lastComponent := slashIdx == -1
-	if lastComponent {
-		slashIdx = len(pattern)
+	separatorAdj := utf8.RuneLen(separator)
+
+	patIdx := indexRuneWithEscaping(pattern, '/')
+	lastPat := patIdx == -1
+	if lastPat {
+		patIdx = len(pattern)
 	}
-	if pattern[:slashIdx] == "**" {
+	if pattern[:patIdx] == "**" {
 		// if our last pattern component is a doublestar, we're done -
 		// doublestar will match any remaining name components, if any.
-		if lastComponent {
+		if lastPat {
 			return true, nil
 		}
 
 		// otherwise, try matching remaining components
-		for nameIdx := 0; nameIdx < nameLen; nameIdx++ {
-			if m, _ := doMatching(pattern[slashIdx+1:], nameComponents[nameIdx:]); m {
+		nameIdx := 0
+		patIdx += 1
+		for {
+			if m, _ := doMatching(pattern[patIdx:], name[nameIdx:], separator); m {
 				return true, nil
 			}
+
+			nextNameIdx := 0
+			if nextNameIdx = indexRuneWithEscaping(name[nameIdx:], separator); nextNameIdx == -1 {
+				break
+			}
+			nameIdx += separatorAdj + nextNameIdx
 		}
 		return false, nil
 	}
 
+	nameIdx := indexRuneWithEscaping(name, separator)
+	lastName := nameIdx == -1
+	if lastName {
+		nameIdx = nameLen
+	}
+
 	var matches []string
-	matches, err = matchComponent(pattern, nameComponents[0])
+	matches, err = matchComponent(pattern, name[:nameIdx])
 	if matches == nil || err != nil {
 		return
 	}
-	if len(matches) == 0 && nameLen == 1 {
+	if len(matches) == 0 && lastName {
 		return true, nil
 	}
 
-	if nameLen > 1 {
+	if !lastName {
+		nameIdx += separatorAdj
 		for _, alt := range matches {
-			matched, err = doMatching(alt, nameComponents[1:])
+			matched, err = doMatching(alt, name[nameIdx:], separator)
 			if matched || err != nil {
 				return
 			}
@@ -311,6 +288,11 @@ func doMatching(pattern string, nameComponents []string) (matched bool, err erro
 // Note: this is meant as a drop-in replacement for filepath.Glob().
 //
 func Glob(pattern string) (matches []string, err error) {
+	return GlobOS(StandardOS, pattern)
+}
+
+// GlobOS is like Glob except that it operates on vos.
+func GlobOS(vos OS, pattern string) (matches []string, err error) {
 	if len(pattern) == 0 {
 		return nil, nil
 	}
@@ -323,7 +305,7 @@ func Glob(pattern string) (matches []string, err error) {
 			return nil, ErrBadPattern
 		}
 		for _, o := range options {
-			m, e := Glob(o + pattern[endOptions+1:])
+			m, e := GlobOS(vos, o+pattern[endOptions+1:])
 			if e != nil {
 				return nil, e
 			}
@@ -341,15 +323,15 @@ func Glob(pattern string) (matches []string, err error) {
 	isWindowsUNC := strings.HasPrefix(volumeName, `\\`)
 	if isWindowsUNC || isAbs {
 		startIdx := len(volumeName) + 1
-		return doGlob(fmt.Sprintf("%s%s", volumeName, string(os.PathSeparator)), filepath.ToSlash(pattern[startIdx:]), matches)
+		return doGlob(vos, fmt.Sprintf("%s%s", volumeName, string(vos.PathSeparator())), filepath.ToSlash(pattern[startIdx:]), matches)
 	}
 
 	// otherwise, it's a relative pattern
-	return doGlob(".", filepath.ToSlash(pattern), matches)
+	return doGlob(vos, ".", filepath.ToSlash(pattern), matches)
 }
 
 // Perform a glob
-func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
+func doGlob(vos OS, basedir, pattern string, matches []string) (m []string, e error) {
 	m = matches
 	e = nil
 
@@ -368,7 +350,7 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 	}
 
 	// Lstat will return an error if the file/directory doesn't exist
-	fi, err := os.Lstat(basedir)
+	fi, err := vos.Lstat(basedir)
 	if err != nil {
 		return
 	}
@@ -380,9 +362,10 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 	}
 
 	// otherwise, we need to check each item in the directory...
+
 	// first, if basedir is a symlink, follow it...
 	if (fi.Mode() & os.ModeSymlink) != 0 {
-		fi, err = os.Stat(basedir)
+		fi, err = vos.Stat(basedir)
 		if err != nil {
 			return
 		}
@@ -393,14 +376,13 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 		return
 	}
 
-	// read directory
-	dir, err := os.Open(basedir)
+	files, err := filesInDir(vos, basedir)
 	if err != nil {
 		return
 	}
-	defer dir.Close()
 
-	files, _ := dir.Readdir(-1)
+	sort.Slice(files, func(i, j int) bool { return files[i].Name() < files[j].Name() })
+
 	slashIdx := indexRuneWithEscaping(pattern, '/')
 	lastComponent := slashIdx == -1
 	if lastComponent {
@@ -411,7 +393,7 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 		for _, file := range files {
 			// if symlink, we may want to follow
 			if (file.Mode() & os.ModeSymlink) != 0 {
-				file, err = os.Stat(filepath.Join(basedir, file.Name()))
+				file, err = vos.Stat(filepath.Join(basedir, file.Name()))
 				if err != nil {
 					continue
 				}
@@ -422,7 +404,7 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 				if lastComponent {
 					m = append(m, filepath.Join(basedir, file.Name()))
 				}
-				m, e = doGlob(filepath.Join(basedir, file.Name()), pattern, m)
+				m, e = doGlob(vos, filepath.Join(basedir, file.Name()), pattern, m)
 			} else if lastComponent {
 				// if the pattern's last component is a doublestar, we match filenames, too
 				m = append(m, filepath.Join(basedir, file.Name()))
@@ -447,7 +429,7 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 				m = append(m, filepath.Join(basedir, file.Name()))
 			} else {
 				for _, alt := range match {
-					m, e = doGlob(filepath.Join(basedir, file.Name()), alt, m)
+					m, e = doGlob(vos, filepath.Join(basedir, file.Name()), alt, m)
 				}
 			}
 		}
@@ -455,6 +437,25 @@ func doGlob(basedir, pattern string, matches []string) (m []string, e error) {
 	return
 }
 
+func filesInDir(vos OS, dirPath string) (files []os.FileInfo, e error) {
+	dir, err := vos.Open(dirPath)
+	if err != nil {
+		return nil, nil
+	}
+	defer func() {
+		if err := dir.Close(); e == nil {
+			e = err
+		}
+	}()
+
+	files, err = dir.Readdir(-1)
+	if err != nil {
+		return nil, nil
+	}
+
+	return
+}
+
 // Attempt to match a single path component with a pattern. Note that the
 // pattern may include multiple components but that the "name" is just a single
 // path component. The return value is a slice of patterns that should be
@@ -498,6 +499,7 @@ func matchComponent(pattern, name string) ([]string, error) {
 				if m, e := matchComponent(pattern[patIdx:], name[nameIdx:]); m != nil || e != nil {
 					return m, e
 				}
+				_, nameAdj = utf8.DecodeRuneInString(name[nameIdx:])
 			}
 			return nil, nil
 		} else if patRune == '[' {
